/*-----------------------------------------------------------------------------

 2002, Jeff M Soft und Hardware GmbH, All Rights Reserved

-----------------------------------------------------------------------------*/
/*
  BK_Reverb   7-29-06   Butch Kratzer --- bkratzer@comcast.net
  http://deztex.home.comcast.net
*/

#include "AllFilter.h"
#include "CombFilter.h"
#include "Module.h"
#include "SEMod_struct.h"
#include "SEPin.h"
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#define PN_IN    0
#define PN_REVERB_TIME      1 // feedback
#define PN_REVERB_AMOUNT    2
#define PN_OUT   3


// The most important part, processing the audio
void Module::sub_process(long buffer_offset, long sampleFrames )
{
	// assign some pointers to your in/output buffers. usually blocks (array) of 96 samples
    in  = buffer_offset + in_buffer;
	out = buffer_offset + out_buffer;

    while(--sampleFrames >= 0) // sampleFrames = how many samples to process (can vary). repeat (loop) that many times
    {
    tcount++;
	 
    vout=*in;
     
    if(Reverb_On)
	{
		pd1 = prebuffer[0][precount[0]];
		prebuffer[0][precount[0]++] = vout/2;
		if(precount[0]>=prebuffer_size[0]) precount[0]=0;		

		pd2 = prebuffer[1][precount[1]];
		prebuffer[1][precount[1]++] = pd1;
		precount[1] %= prebuffer_size[1];		
		if(precount[1]>=prebuffer_size[1]) precount[1]=0;		


        groupa = comba[0].process(pd1);
        groupa = groupa+comba[1].process(pd1);
        groupa = groupa+comba[2].process(pd1);
        groupa = groupa+comba[3].process(pd1);

        groupb = combb[0].process(pd2);
        groupb = groupb+combb[1].process(pd2);
        groupb = groupb+combb[2].process(pd2);
        groupb = groupb+combb[3].process(pd2);


        wetouta = allpassa[0].process(groupa);
        wetouta = allpassb[0].process(wetouta);

        wetoutb = allpassa[1].process(groupb);
        wetoutb = allpassb[1].process(wetoutb);

	   
	   vout=((wetouta+wetoutb)*Reverb_Amount+vout);
	
	}

	
	*out=vout;

	
	if(vout>.001f) tcount=0;


     in++; out++;
	}
    if(tcount>300) SetSleepMode(ST_STATIC);
}

// every module has an output buffer of approx 100 samples.
// before deactivating module, need to fill that buffer with silence.
// "static_count" is counting out those 100 samples.
void Module::sub_process_static(long buffer_offset, long sampleFrames )
{
	sub_process(buffer_offset, sampleFrames);
	static_count = static_count - sampleFrames;
	if( static_count <= 0 )
	{
		CallHost(seaudioMasterSleepMode);
	}
}

Module::Module(seaudioMasterCallback2 seaudioMaster, void *p_resvd1) : SEModule_base(seaudioMaster, p_resvd1)
{
}


Module::~Module()
{
	// This is where you free any memory/resources your module has created

	    delete [] prebuffer[0];
		delete [] prebuffer[1];
}

// describe your module
bool Module::getModuleProperties (SEModuleProperties* properties)
{
char data1[1000],data2[255];

sprintf(data1,"This module developed by Butch Kratzer\n\n");
sprintf(data2,"bkratzer@comcast.net");
strcat(data1,data2);
	
	// describe the plugin, this is the name the end-user will see.
	properties->name = "BK_Reverb";								// !!TODO!!

	// return a unique string 32 characters max
	// if posible include manufacturer and plugin identity
	// this is used internally by SE to identify the plug.
	// No two plugs may have the same id.
	properties->id = "BK_Reverb";						// !!TODO!!

	// Info, may include Author, Web page whatever
	properties->about = data1;								// !!TODO!!
	return true;
}

// describe the pins (plugs and parameters)
bool Module::getPinProperties (long index, SEPinProperties* properties)
{
	switch( index )								// !!TODO!! list your in / out plugs
	{
		// typical input plug (inputs are listed first)


	case 0:
		properties->name				= "In";
		properties->variable_address	= &in_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;
	
	case 1:
		properties->name				= "Reverb Time";
		properties->variable_address	= &reverb_time_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;

	case 2:
		properties->name				= "Reverb Amount";
		properties->variable_address	= &reverb_amount_buffer;
		properties->direction			= DR_IN;
		properties->datatype			= DT_FSAMPLE;
		properties->default_value		= "0";
		break;

    case 3:
		properties->name				= "Out";
		properties->variable_address	= &out_buffer;
		properties->direction			= DR_OUT;
		properties->datatype			= DT_FSAMPLE;
		break;

	
	default:
		return false; // host will ask for plugs 0,1,2,3 etc. return false to signal when done
	};

	return true;
}

// An input plug has changed value
void Module::OnPlugStateChange(SEPin *pinptr)
{
int pin;
     

     pin=pinptr->getPinID(); 	

	 if(pin==PN_IN)
	 {
		 tcount=0;
		 SetSleepMode(ST_RUN);
		 
	 }

    if(pin==PN_REVERB_AMOUNT) Reverb_UpdateAmount();
	if(pin==PN_REVERB_TIME)   Reverb_UpdateTime();
}

void Module::open()
{
	// call the base class
	SEModule_base::open();

	// choose which function is used to process audio
	SET_PROCESS_FUNC(Module::sub_process);


    StartUp();

}

void Module::StartUp()
{
int ctr;
SamplesPerSec=getSampleRate();

tcount=0;

Reverb_CreatePreBuffer();
Reverb_UpdateTime();
Reverb_UpdateAmount();


	comba[0].setbuffer(bufcomba1,1116);
	comba[1].setbuffer(bufcomba2,1277);
	comba[2].setbuffer(bufcomba3,1422*2);
	comba[3].setbuffer(bufcomba4,1577*2);

	combb[0].setbuffer(bufcombb1,1188);
	combb[1].setbuffer(bufcombb2,1356);
	combb[2].setbuffer(bufcombb3,1491*2);
	combb[3].setbuffer(bufcombb4,1617*3);

	
	allpassa[0].setbuffer(bufallpassa1,556*2);
    allpassa[1].setbuffer(bufallpassa2,341*2);

	allpassb[0].setbuffer(bufallpassb1,441*2);
    allpassb[1].setbuffer(bufallpassb2,225*2);

	for(ctr=0; ctr<=3; ctr++)
	{
		comba[ctr].mute();
		combb[ctr].mute();
		comba[ctr].setdamp(0.25f);
		combb[ctr].setdamp(0.25f); 
	}

    for(ctr=0; ctr<=1; ctr++)
	{  
		allpassa[ctr].mute();
		allpassb[ctr].mute();
	}

SetSleepMode(ST_RUN);
}

void Module::SetSleepMode(state_type out_stat)
{

	getPin(PN_OUT)->TransmitStatusChange( SampleClock(), out_stat);


	// setup up 'sleep mode' or not
	if( out_stat < ST_RUN )
	{
		static_count = getBlockSize();
		SET_PROCESS_FUNC(Module::sub_process_static);
		sleepmode=true;
	}
	else
	{
		SET_PROCESS_FUNC(Module::sub_process);
		sleepmode=false;
	}

}


void Module::Reverb_UpdateAmount()
{
	
	Reverb_Amount=getPin(PN_REVERB_AMOUNT)->getValue();
    if(Reverb_Amount>=0.05) Reverb_On=true; else Reverb_On=false;

}


void Module::Reverb_UpdateTime()
{
int ctr;
	
	Reverb_Time=getPin(PN_REVERB_TIME)->getValue()*0.85f;

    prebuffer_size[0] = (unsigned long) ( SamplesPerSec * Reverb_Time/10 );
    prebuffer_size[1] = (unsigned long) ( SamplesPerSec * Reverb_Time/15 );

	for(ctr=0; ctr<=3; ctr++)
	{
		comba[ctr].setfeedback(Reverb_Time);
		combb[ctr].setfeedback(Reverb_Time);
	
	}

    for(ctr=0; ctr<=1; ctr++)
	{  
		allpassa[ctr].setfeedback(Reverb_Time);
		allpassb[ctr].setfeedback(Reverb_Time);
	}


}

void Module::Reverb_CreatePreBuffer()
{
int ctr;
long size;

	// free the memory currently used by the buffer (if any)
  for(ctr=0; ctr<=1; ctr++)
  {

	if( prebuffer[ctr] != NULL )
	{
		delete [] prebuffer[ctr];
		prebuffer[ctr] = NULL;
	}


	size = (long)SamplesPerSec;

	prebuffer[ctr] = new float[size];
	memset(prebuffer[ctr], 0, size * sizeof(float) ); // clear buffer

	precount[ctr] = 0;
  }
}

